(** Copyright (c) 2018 by Swirlds, Inc.
   Implemented by Karl Crary.
*)

Require Import Nat.
Require Import Bool.
Require Import Omega.
Require Import List.
Require Import Permutation.

Require Import Tact.
Require Import Decide.
Require Import Relation.
Require Import Majority.
Require Import Cardinality.
Require Import HashgraphFacts.
Require Import Sees.
Require Import Round.
Require Import Progress.
Require Import Vote.
Require Import Decision.
Require Import Consensus.
Require Import Famous.
Require Import Received.
Require Import Median.
Require Import Order.


Lemma stsees_impl_famous :
  forall s i x y,
    first_regular >= 2
    -> member (global s) y
    -> rwitness i x
    -> (forall j z, parent z y -> round j z -> j <= i)
    -> stsees x y
    -> famous s (global s) x.
Proof.
intros s i x y Hd2 Hsy Hwitx Hparentround Hxy.
so (all_rounds_inhabited _ (S first_regular + i) (global_nonterminating s)) as (z & Hwitz & Hsz).
exists z.
split; auto.
apply (unanimity_impl_decision s (global s) x z i (first_regular + i) true); auto.
  {
  exact (Hwitx andel).
  }

2:{
  omega.
  }

2:{
  replace (S (first_regular + i) - i) with (S first_regular) by omega.
  so coin_freq_min.
  rewrite -> Nat.mod_small; omega.
  }
clear z Hwitz Hsz.
intros w Hsw Hwitw.
eapply vote_first; eauto; try omega.
split.
2:{
  intros _.
  exact I.
  }
intros _.
(** The heart of the proof starts here. *)
destruct Hxy as (y' & Hyy & Hmaj1).
exploit (stsees_many_witnesses (pred (first_regular + i)) w) as Hmaj2.
  {
  force_exact (Hwitw andel).
  f_equal.
  omega.
  }
so (supermajority_intersect_3 _#5 eq_peer_decide supermajority_honest Hmaj1 Hmaj2) as (a & Hhonest & Ha1 & Ha2).
destruct Ha1 as (u & Heqcru & Hxu & Huy).
destruct Ha2 as (v & Heqcrv & Hvw & Hwitv).
apply (star_plus_trans _#3 u).
  {
  apply sees_impl_ancestor; auto.
  }
apply (star_plus_trans _#3 v).
2:{
  apply stsees_impl_strict_ancestor; auto.
  }
apply self_ancestor_impl_ancestor.
destruct (eq_event_decide v u) as [Heq | Hneqvu].
  {
  subst v.
  apply star_refl.
  }
rewrite <- Heqcru in Hhonest.
rewrite <- Heqcrv in Heqcru.
exploit (self_ancestor_total (global s) u v) as H; auto.
  {
  apply (world_closed _ _ y); auto.
  eapply star_trans; eauto.
  apply sees_impl_ancestor; auto.
  }

  {
  apply (world_closed _ _ w); auto.
  apply stsees_impl_ancestor; auto.
  }
destruct H as [| Hvu]; auto.
exfalso.
so (eq_event_decide u y) as [Heq | Hnequy].
  {
  subst u.
  so (plus_plusr _#4 (star_neq_plus _#4 Hvu Hneqvu)) as (z & Hvz & Hzy).
  so (round_defined z) as (j & Hroundz).
  so (Hparentround _ _ (self_parent_impl_parent _ _ Hzy) Hroundz) as Hji.
  so (round_mono _#4 (self_ancestor_impl_ancestor _ _ Hvz) (Hwitv andel) Hroundz).
  omega.
  }

  {
  so (plus_plusr _#4 (star_neq_plus _#4 (star_trans _#5 (sees_impl_ancestor _ _ Huy) Hyy) Hnequy)) as (z & Huz & Hzy).
  so (round_defined z) as (j & Hroundz).
  so (Hparentround _ _ Hzy Hroundz) as Hji.
  so (round_mono _#4 (star_trans _#5 (self_ancestor_impl_ancestor _ _ Hvu) Huz) (Hwitv andel) Hroundz).
  omega.
  }
Qed.


Theorem fairness :
  forall s i,
    first_regular >= 2
    -> supermajority
         (fun a => exists w, creator w = a /\ fwitness s (global s) i w)
         every.
Proof.
intros s i Hd2.
assert (exists y,
          member (global s) y
          /\ rwitness (S i) y
          /\ forall j z, parent z y -> round j z -> j <= i).
  {
  so (all_rounds_inhabited _ (S i) (global_nonterminating s)) as (y & Hwity & Hsy).
  revert Hwity Hsy.
  wfinduct y using strict_ancestor_well_founded.
  intros y IH Hwity Hsy.
  so (initial_decide y) as [Hinit | (u & v & Hparents)].
    {
    so (round_fun _#3 (Hwity andel) (round_initial Hinit)) as H.
    discriminate H.
    }
  so (round_defined v) as (j & Hroundv).
  so (le_dec j i) as [Hle | Hnle].
    {
    exists y.
    do2 2 split; auto.
    intros k z Hzy Hroundz.
    invert Hzy.
      {
      intros ? Hparents'.
      so (parents_fun _#5 Hparents Hparents') as (<- & <-).
      so (strict_self_ancestor_witness _#4 (Hwity ander) (plus_one _#4 (ex_intro _ _ Hparents)) Hroundz (Hwity andel)).
      omega.
      }

      {
      intros ? Hparents'.
      so (parents_fun _#5 Hparents Hparents') as (<- & <-).
      so (round_fun _#3 Hroundv Hroundz); subst k.
      auto.
      }
    }

    {
    so (ancestor_witness _#3 (le_refl _) Hroundv) as (w & Hwv & Hwitw).
    assert (w @ y) as Hwy.
      {
      eapply star_plus_trans; eauto.
      apply plus_one.
      eapply parent2; eauto.
      }
    apply (IH w); auto.
      {
      force_exact Hwitw.
      f_equal.
      so (round_mono _#4 (plus_star _#4 Hwy) (Hwitw andel) (Hwity andel)).
      omega.
      }

      {
      apply (world_closed _ _ y); auto.
      apply plus_star; auto.
      }
    }
  }
destruct H as (y & Hsy & Hwity & Hparentround).
so (stsees_many_witnesses _ _ (Hwity andel)) as Hmaj.
eapply supermajority_enlarge; eauto.
  {
  intros a Ha.
  destruct Ha as (x & Heqcr & Hxy & Hwitx).
  exists x.
  split; auto.
  split; auto.
  eapply stsees_impl_famous; eauto.
  }

  {
  intros a _.
  apply (dec_exists_finite _ _ (fwitness s (global s) i)).
    {
    intros x (H1 & H2).
    auto.
    }

    { apply fwitness_finite_global. }

    {
    intros x.
    apply dec_and.
      {
      apply eq_peer_decide.
      }

      {
      apply finite_decide.
        {
        exact eq_event_decide.
        }

        {
        apply fwitness_finite_global.
        }
      }
    }
  }
Qed.


Lemma ufwitness_fun :
  forall s W i x y,
    creator x = creator y
    -> ufwitness s W i x
    -> ufwitness s W i y
    -> x = y.
Proof.
intros s W i x y Heqcr Hx Hy.
apply (ord_antisym _ _ textorder_order).
  {
  apply (Hx ander); auto.
  exact (Hy andel).
  }

  {
  apply (Hy ander); auto.
  exact (Hx andel).
  }
Qed.


Lemma list_minimum :
  forall (T : Type) (R : T -> T -> Prop) l,
    transitive T R
    -> (forall x y, R x y \/ R y x)
    -> l = nil \/ exists x, In x l /\ forall y, In y l -> R x y.
Proof.
intros T R l Htrans Htotal.
induct l.

(** nil *)
{
left; auto.
}

(** cons *)
{
intros x l IH.
right.
destruct IH as [-> | (y & Hy)].
  {
  exists x.
  split.
    { left; auto. }
  intros y Hy.
  destruct Hy as [Heq | H].
    {
    subst y.
    destruct (Htotal x x); auto.
    }

    {
    destruct H.
    }
  }

  {
  destruct Hy as (Hin & Hy).
  so (Htotal x y) as [Hxy | Hyx].
    {
    exists x.
    split.
      { left; auto. }
    intros z Hz.
    destruct Hz as [Heq | Hz].
      {
      subst z.
      destruct (Htotal x x); auto.
      }

      {
      eapply Htrans; eauto.
      }
    }

    {
    exists y.
    split.
      { right; auto. }
    intros z Hz.
    destruct Hz as [Heq | Hz].
      {
      subst z.
      auto.
      }

      {
      apply Hy; auto.
      }
    }
  }
}
Qed.


Lemma obtain_ufwitness :
  forall s W i x,
    finite (fwitness s W i)
    -> fwitness s W i x
    -> exists y,
         creator x = creator y
         /\ ufwitness s W i y.
Proof.
intros s W i x Hfin Hfwit.
exploit (finite_subset _ (fun z => creator x = creator z /\ fwitness s W i z) (fwitness s W i)) as Hfin'; auto.
  {
  intros z (_ & H); auto.
  }

  {
  intros z.
  apply dec_and.
    { apply eq_peer_decide. }

    {
    apply finite_decide; auto.
    exact eq_event_decide.
    }
  }
destruct Hfin' as (l & Hl).
so (list_minimum _ _ l (ord_trans _ _ textorder_order) textorder_total) as [-> | (y & Hin & Hy)].
  {
  exfalso.
  refine (Hl x andel _).
  split; auto.
  }
so (Hl y ander Hin) as (Heqcr & Hfwit').
exists y.
split; auto.
split; auto.
intros z Hz Heqcr'.
apply Hy.
apply Hl.
split; auto.
transitivity (creator y); auto.
Qed.


Lemma earliest_observer_fun :
  forall x y y' z,
    earliest_observer x y z
    -> earliest_observer x y' z
    -> y = y'.
Proof.
intros x y z w Hy Hz.
destruct Hy as (Hxy & Hy & Hyw).
destruct Hz as (Hxz & Hz & Hzw).
so (star_starr _#4 Hyw) as H; renameover H into Hyw.
so (star_starr _#4 Hzw) as H; renameover H into Hzw.
revert Hyw Hzw.
wfinduct w using parent_well_founded.
intros w IH Hyw Hzw.
invertc Hyw.

(** refl *)
{
intros ->.
invertc Hzw.
  (** refl *)
  {
  intros ->.
  reflexivity.
  }

  (** step *)
  {
  intros z' Hzz' Hz'w.
  exfalso.
  apply (Hy z'); auto.
  eapply star_trans; eauto.
  apply self_ancestor_impl_ancestor.
  apply starr_star; auto.
  }
}

(** step *)
{
intros y' Hyy' Hy'w.
invertc Hzw.
  (** refl *)
  {
  intros ->.
  exfalso.
  apply (Hz y'); auto.
  apply (star_trans _#3 y); eauto.
  apply self_ancestor_impl_ancestor.
  apply starr_star; auto.
  }

  (** step *)
  {
  intros z' Hzz' Hz'w.
  destruct Hy'w as (u & Hparents).
  destruct Hz'w as (u' & Hparents').
  so (parents_fun _#5 Hparents Hparents') as (<- & _).
  apply (IH y'); auto.
  eapply parent1; eauto.
  }
}
Qed.


Lemma earliest_observer_defined :
  forall x z,
    x @= z
    -> exists y,
         earliest_observer x y z.
Proof.
intros x z Hxz.
revert x Hxz.
wfinduct z using parent_well_founded.
intros z IH x Hxz.
so (initial_decide z) as [Hinitial | (u & v & Hparents)].
  {
  so (star_plus _#4 Hxz) as [Heq | Hxz'].
    {
    subst z.
    exists x.
    do2 2 split.
      { apply star_refl. }

      {
      intros w Hwx Hxw.
      apply (strict_ancestor_irreflex x).
      eapply star_plus_trans; eauto.
      apply plus_one.
      apply self_parent_impl_parent; auto.
      }
      
      { apply star_refl. }
    }

    {
    exfalso.
    so (plus_plusr _#4 Hxz') as (u & _ & Huz).
    eapply initial_no_parent; eauto.
    }
  }
so (ancestor_decide x u) as [Hxu | Hnxu].
  {
  so (IH u (parent1 Hparents) x Hxu) as (y & Hearliest).
  exists y.
  destruct Hearliest as (Hxy & Hfirst & Hyu).
  do2 2 split; auto.
  eapply star_trans; eauto.
  apply star_one.
  exists v; auto.
  }

  {
  exists z.
  do2 2 split; auto.
    2:{ apply star_refl. }
  intros u' Hu'z Hxy.
  destruct Hu'z as (v' & Hparents').
  so (parents_fun _#5 Hparents Hparents') as (<- & <-).
  contradiction.
  }
Qed.


Corollary honest_timestamp :
  forall s x t,
    first_regular >= 2
    -> time_received s (global s) x t
    -> exists i y y' w w',
         received s (global s) i x
         /\ honest (creator y)
         /\ honest (creator y')
         /\ ufwitness s (global s) i w
         /\ ufwitness s (global s) i w'
         /\ earliest_observer x y w
         /\ earliest_observer x y' w'
         /\ timestamp y <= t <= timestamp y'.
Proof.
intros s x t Hd2 Htr.
destruct Htr as (i & l & Hrec & Hdist & Hl & ->).
exploit (majority_median _ (fun y => honest (creator y) /\ In y l) timestamp l) as H.
  { exact eq_event_decide. }

  { exact Hdist. }

  {
  so (fairness s i Hd2) as Hmaj.
  so (supermajority_intersect_2' _#4 eq_peer_decide supermajority_honest Hmaj) as (_ & Hmaj').
  destruct Hmaj' as (m & n & _ & Hcardhonest & Hcardall & Hsize).
  exists m, n.
  so (cardinality_In _ l Hdist) as Hcardl.
  eassert _ as H; [refine (cardinality_subset _ (fun y => honest (creator y) /\ In y l) _ _ _ _ Hcardl) |].
    {
    intros y Hy.
    apply dec_and.
      { apply honest_decide. }

      { left; auto. }
    }

    {
    intros y (_ & H); auto.
    }
  destruct H as (m' & Hcardhonest' & _).
  assert (m = m').
    {
    refine (cardinality_corr_eq _#4 (fun a y => creator y = a) _ _ _ _ _ _ Hcardhonest Hcardhonest').
      {
      intros a a' y _ _ _ Heq Heq'.
      transitivity (creator y); auto.
      }

      {
      intros a y y' _ (_ & Hy) (_ & Hy') Heqcr Heqcr'.
      rewrite <- Heqcr' in Heqcr.
      so (Hl _ ander Hy) as (w & Hw & Hearliest).
      so (Hl _ ander Hy') as (w' & Hw' & Hearliest').
      assert (creator w = creator w') as Heqcrw.
        {
        so (self_ancestor_creator _ _ (earliest_observer_impl_self_ancestor _#3 Hearliest)).
        so (self_ancestor_creator _ _ (earliest_observer_impl_self_ancestor _#3 Hearliest')).
        transitivity (creator y); auto.
        transitivity (creator y'); auto.
        }
      so (ufwitness_fun _#5 Heqcrw Hw Hw'); subst w'.
      eapply earliest_observer_fun; eauto.
      }
      
      {
      intros a (Hhonest & (w & Heqcr & Hw)).
      exploit (obtain_ufwitness s (global s) i w) as H; auto.
        {
        apply fwitness_finite_global.
        }
      destruct H as (v & Heqcrv & Hv).
      so (earliest_observer_defined _ _ (Hrec andel ander _ Hv)) as (z & Hearliest).      
      exists z.
      do2 2 split.
        {
        transitivity (creator w); auto.
        transitivity (creator v); auto.
        apply self_ancestor_creator.
        eapply earliest_observer_impl_self_ancestor; eauto.
        }

        {
        force_exact Hhonest.
        f_equal.
        transitivity (creator w); auto.
        transitivity (creator v); auto.
        symmetry.
        apply self_ancestor_creator.
        eapply earliest_observer_impl_self_ancestor; eauto.
        }

        {
        apply Hl.
        exists v.
        auto.
        }
      }

      {
      intros y (Hhonest & Hy).
      exists (creator y).
      do2 2 split; auto.
      so (Hl _ ander Hy) as (w & Hw & Hearliest).
      exists w.
      split.
        {
        symmetry.
        apply self_ancestor_creator.
        eapply earliest_observer_impl_self_ancestor; eauto.
        }

        {
        exact (Hw andel).
        }
      }
    }
  subst m'.
  assert (n = length l) as Heq.
    {
    refine (cardinality_corr_eq _#4 (fun a y => creator y = a) _ _ _ _ _ _ Hcardall Hcardl).
      {
      intros a a' y _ _ _ Heq Heq'.
      transitivity (creator y); auto.
      }

      {
      intros a y y' _ Hy Hy' Heqcr Heqcr'.
      rewrite <- Heqcr' in Heqcr.
      so (Hl _ ander Hy) as (w & Hw & Hearliest).
      so (Hl _ ander Hy') as (w' & Hw' & Hearliest').
      assert (creator w = creator w') as Heqcrw.
        {
        so (self_ancestor_creator _ _ (earliest_observer_impl_self_ancestor _#3 Hearliest)).
        so (self_ancestor_creator _ _ (earliest_observer_impl_self_ancestor _#3 Hearliest')).
        transitivity (creator y); auto.
        transitivity (creator y'); auto.
        }
      so (ufwitness_fun _#5 Heqcrw Hw Hw'); subst w'.
      eapply earliest_observer_fun; eauto.
      }
      
      {
      intros a (w & Heqcr & Hw).
      exploit (obtain_ufwitness s (global s) i w) as H; auto.
        {
        apply fwitness_finite_global.
        }
      destruct H as (v & Heqcrv & Hv).
      so (earliest_observer_defined _ _ (Hrec andel ander _ Hv)) as (z & Hearliest).      
      exists z.
      split.
        {
        transitivity (creator w); auto.
        transitivity (creator v); auto.
        apply self_ancestor_creator.
        eapply earliest_observer_impl_self_ancestor; eauto.
        }
      apply Hl.
      exists v.
      auto.
      }

      {
      intros y Hy.
      exists (creator y).
      split; auto.
      so (Hl _ ander Hy) as (w & Hw & Hearliest).
      exists w.
      split.
        {
        symmetry.
        apply self_ancestor_creator.
        eapply earliest_observer_impl_self_ancestor; eauto.
        }

        {
        exact (Hw andel).
        }
      }
    }
  rewrite <- Heq in Hcardl.
  do2 3 split; auto.
  intros y (_ & H); auto.
  }
destruct H as (y & y' & (Hhonesty & Hy) & (Hhonesty' & Hy') & Hstamplo & Hstamphi).
so (Hl _ ander Hy) as (w & Hw & Hearliest).
so (Hl _ ander Hy') as (w' & Hw' & Hearliest').
exists i, y, y', w, w'.
do2 8 split; auto.
Qed.
